/* wydawca - automatic release submission daemon
   Copyright (C) 2007, 2009 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "mail.h"
#include "argmatch.h"
#include "version-etc.h"

char *conffile = SYSCONFDIR "/wydawca.rc" ;
int debug_level;
int dry_run_mode;
int log_to_stderr = -1; /* -1 means autodetect */
int log_facility = LOG_LOCAL1;
char *syslog_tag = "wydawca";
int syslog_include_prio; /* syslog messages include priority */
unsigned long print_stats;         /* Print final statistics output */
time_t file_sweep_time = 0;
char *tar_command_name = "tar";
int enable_sendfile = 1;  /* Use sendfile by default */
int archive_signatures = 1; /* Archive sig files by default */
int lint_mode = 0;
int preprocess_only = 0;
  
unsigned wydawca_stat[MAX_STAT];

void
usage ()
{
  printf ("usage: wydawca [OPTIONS]\n");
  printf ("wydawca synchronizes files from a set of upload directories with the\n"
	  "corresponding distribution sites.\n");
  printf ("\n");
  printf ("OPTIONS are:\n");
  printf ("  -c, --config-file=FILE         use FILE instead of the default configuration\n"
	  "                                 file\n");
  printf ("      --cron, --syslog           log to syslog\n");
  printf ("  -d, --debug                    increase debugging level by 1\n");
  printf ("  -e, --stderr                   log to stderr\n");
  printf ("  -I, --include-directory[=DIR]  set default include directory\n");
  printf ("  -n, --dry-run                  do nothing, print almost everything;\n"
	  "                                 implies `--debug --stderr'; use additional\n"
	  "                                 `--debug' options to get even more info\n");
  printf ("  -t, --lint                     parse configuration file and exit\n");
  printf ("  -h, --help                     print this help list\n");
  printf ("  -v, --version                  print program version and exit\n");

  printf ("\nThe defaults are:\n");
  printf ("Configuration file: %s\n", conffile);
  printf ("Syslog facility: %s\n",
	  mu_syslog_facility_to_string (log_facility));
  printf ("Syslog tag: %s\n", syslog_tag);
  printf ("\n");
  printf ("Report bugs to <%s>\n", PACKAGE_BUGREPORT);
}


/* Logging */
void
syslog_printer (int prio, const char *fmt, va_list ap)
{
  if (syslog_include_prio)
    {
      static char *fmtbuf;
      static size_t fmtsize;
      const char *p = mu_syslog_priority_to_string (prio);
      size_t size = strlen (p) + 3 + strlen (fmt) + 1;

      if (size > fmtsize)
	{
	  fmtsize = size;
	  fmtbuf = x2realloc (fmtbuf, &fmtsize);
	}
      sprintf (fmtbuf, "[%s] %s", p, fmt);
      fmt = fmtbuf;
    }
  
#if HAVE_VSYSLOG
  vsyslog (prio, fmt, ap);
#else
  char buf[128];
  vsnprintf (buf, sizeof buf, fmt, ap);
  syslog (prio, "%s", buf);
#endif
}

void
stderr_printer (int prio, const char *fmt, va_list ap)
{
  const char *p = mu_syslog_priority_to_string (prio);
  fprintf (stderr, "%s: ", program_name);

  if (p)
    fprintf (stderr, "[%s] ", p);

  vfprintf (stderr, fmt, ap);
  fputc ('\n', stderr);
}


static void (*log_printer) (int prio, const char *fmt, va_list ap) =
         stderr_printer;

void
logmsg (int prio, char *fmt, ...)
{
  va_list ap;

  switch (prio)
    {
    case LOG_EMERG:
    case LOG_ALERT:
    case LOG_CRIT:
    case LOG_ERR:
      UPDATE_STATS (STAT_ERRORS);
      break;
      
    case LOG_WARNING:
      UPDATE_STATS (STAT_WARNINGS);
    }
      
  va_start (ap, fmt);
  log_printer (prio, fmt, ap);
  va_end (ap);
}



enum {
  SYSLOG_OPTION = 256,
  CONFIG_HELP_OPTION,
  DUMP_GRAMMAR_TRACE_OPTION,
  DUMP_LEX_TRACE_OPTION
};

struct option options[] = {
  { "config-file", required_argument, NULL, 'c' },
  { "config-help", no_argument, NULL, CONFIG_HELP_OPTION },
  { "dump-grammar-trace", no_argument, NULL, DUMP_GRAMMAR_TRACE_OPTION },
  { "dump-lex-trace", no_argument, NULL, DUMP_LEX_TRACE_OPTION },
  { "debug", no_argument, NULL, 'd' },
  { "stderr", no_argument, NULL, 'e' },
  { "syslog", no_argument, NULL, SYSLOG_OPTION },
  { "cron", no_argument, NULL, SYSLOG_OPTION },
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'v' },
  { "dry-run", no_argument, NULL, 'n' },
  { "lint", no_argument, NULL, 't' },
  { "include-directory", optional_argument, NULL, 'I' },
  { NULL }
};


static char *stat_name[MAX_STAT] = {
  "errors",
  "warnings",
  "bad signatures",
  "access violation attempts",
  "complete triplets",
  "incomplete triplets",
  "bad triplets",
  "expired triplets",
  "triplet successes",
  "files uploaded",
  "files archived",
  "symlinks created",
  "symlinks removed",
};

static char *stat_kwname[MAX_STAT] = {
  "stat:errors",               
  "stat:warnings",             
  "stat:bad_signatures",       
  "stat:access_violations",    
  "stat:complete_triplets",    
  "stat:incomplete_triplets",  
  "stat:bad_triplets",         
  "stat:expired_triplets",     
  "stat:triplet_success",      
  "stat:uploads",              
  "stat:archives",             
  "stat:symlinks",             
  "stat:rmsymlinks",      
};

int
stat_mask_p (unsigned long mask)
{
  int i;

  for (i = 0; i < MAX_STAT; i++)
    if (wydawca_stat[i] != 0 && (mask && STAT_MASK (i)))
      return 1;
  return 0;
}

static const char *
stat_expand (struct metadef *def, void *data)
{
  char sbuf[INT_BUFSIZE_BOUND (wydawca_stat[0])];
  def->storage = xstrdup (uinttostr (wydawca_stat[(int) def->data], sbuf));
  def->value = def->storage;
  return def->value;
}

void
make_stat_expansion (struct metadef *def)
{
  int i;
  for (i = 0; i < MAX_STAT; i++)
    {
      def[i].kw = stat_kwname[i];
      def[i].value = NULL;
      def[i].storage = NULL;
      def[i].expand = stat_expand;
      def[i].data = (void*) i;
    }
  def[i].kw = NULL;
}

void
logstats ()
{
  int i;

  if (stat_mask_p (print_stats))
    {
      for (i = 0; i < MAX_STAT; i++)
	if (print_stats & STAT_MASK (i))
	  logmsg (LOG_INFO, "%s: %u", stat_name[i], wydawca_stat[i]);
    }
  
  mail_stats ();
}


void
gconf_print_diag (gconf_locus_t *locus, int err, int errcode, const char *msg)
{
  if (locus)
    {
      if (errcode)
	logmsg (err ? LOG_ERR : LOG_WARNING, "%s:%lu: %s: %s",
		locus->file, (unsigned long)locus->line, msg,
		strerror (errcode));
      else
	logmsg (err ? LOG_ERR : LOG_WARNING, "%s:%lu: %s",
		locus->file, (unsigned long)locus->line, msg);
    }
  else
    {
      if (errcode)
	logmsg (err ? LOG_ERR : LOG_WARNING, "%s: %s", msg,
		strerror (errcode));
      else
	logmsg (err ? LOG_ERR : LOG_WARNING, "%s", msg);
    }
}

#include "cmdline.h"
  
int
main (int argc, char **argv)
{
  program_name = argv[0];
  mu_register_all_mailer_formats ();
  config_init ();

  parse_options (argc, argv);
  
  if (argc != optind)
    error (1, 0, "extra command line arguments");

  if (preprocess_only)
    exit (gconf_preproc_run (conffile, gconf_preprocessor));
  
  gconf_parse (conffile);
  
  if (lint_mode)
    exit (0);

  if (log_to_stderr == -1)
    log_to_stderr = isatty (0);

  gconf_log_to_stderr = log_to_stderr;
  if (!log_to_stderr)
    {
      openlog (syslog_tag, LOG_PID, log_facility);
      log_printer = syslog_printer;
    }

  mail_init ();
  
  logmsg (LOG_NOTICE, "wydawca (%s) started", PACKAGE_STRING);
  scan_directories ();

  logmsg (LOG_NOTICE, "wydawca (%s) finished", PACKAGE_STRING);
  logstats ();

  mail_finish ();

  exit (0);
}
